/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2022 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.16.1.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : Initializes and handles the hardware for embOS
*/

#include "RTOS.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ  (24500000uL)
#define OS_TICK_FREQ   (1000u)
#define OS_INT_FREQ    (OS_TICK_FREQ)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_UART
#endif

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #define OS_UART       (0u)
  #define OS_BAUDRATE   (38400u)
  #define OS_UART_FREQ  (OS_TIMER_FREQ)
#endif

/*********************************************************************
*
*       Device specific SFRs
*/
__sfr __no_init volatile unsigned char PCA0MD  @ 0xD9; // PCA Mode Register
__sfr __no_init volatile unsigned char TMR2CN  @ 0xC8; // Timer 2 Control Register
__sfr __no_init volatile unsigned char TMR2RLL @ 0xCA; // Timer 2 Reload Register Low Byte
__sfr __no_init volatile unsigned char TMR2RLH @ 0xCB; // Timer 2 Reload Register High Byte
__sfr __no_init volatile unsigned char TMR2RL  @ 0xCC; // Timer 2 Low Byte
__sfr __no_init volatile unsigned char TMR2RH  @ 0xCD; // Timer 2 High Byte
__sfr __no_init volatile unsigned char IE      @ 0xA8; // Interrupt Enable
__sfr __no_init volatile unsigned char CLKSEL  @ 0xA9; // Clock select
__sfr __no_init volatile unsigned char CKCON   @ 0x8E; // Clock control
__sfr __no_init volatile unsigned char OSCICN  @ 0xB2; // Internal Oscillator Control
__sfr __no_init volatile unsigned char RSTSRC  @ 0xEF; // Reset register

#define WDTE_BIT         (6u)
#define IE_ET2_BIT       (5u)
#define TMR2CN_TR2_BIT   (2u)
#define TMR2_TF2H_MASK   (1u << 7u)

__sfr __no_init volatile unsigned char TCON    @ 0x88; // Timer Control
__sfr __no_init volatile unsigned char TMOD    @ 0x89; // Timer 1 Control Register
__sfr __no_init volatile unsigned char TH1     @ 0x8D; // Timer 1 Reload Register Low Byte
__sfr __no_init volatile unsigned char TL1     @ 0x8B; // Timer 1 Reload Register High Byte
__sfr __no_init volatile unsigned char XBR0    @ 0xE1; // XBAR 0
__sfr __no_init volatile unsigned char XBR2    @ 0xE3; // XBAR 1
__sfr __no_init volatile unsigned char P0MDOUT @ 0xA4; // Port 0 Output Mode Configuration
__sfr __no_init volatile unsigned char SCON0   @ 0x98; // Serial Port 0 Control
__sfr __no_init volatile unsigned char SBUF0   @ 0x99; // Serial (UART0) Port Data Buffer

#define UART0_REN0_MASK  (1u << 4u)
#define UART0_TI0_MASK   (1u << 1u)
#define UART0_RI0_MASK   (1u << 0u)
#define IE_ES0_MASK      (1u << 4u)

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/****************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return 0;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return 0;
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       SysTick_Handler()
*
*  Function description
*    This is the hardware timer exception handler.
*/
#pragma vector = 0x2B
__interrupt void SysTick_Handler(void);
__interrupt void SysTick_Handler(void) {
  TMR2CN &= ~TMR2_TF2H_MASK;
  OS_INT_EnterNestable();
  OS_TICK_Handle();
  OS_INT_LeaveNestable();
}

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
void OS_InitHW(void) {
  OS_INT_IncDI();
  //
  // Inform embOS about the timer settings
  //
  {
    OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_DOWNCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};
    OS_TIME_ConfigSysTimer(&SysTimerConfig);
  }
  //
  // Initialize timer for embOS, assuming clock is already initialized
  //
  TMR2RL  = (0x10000 - (OS_TIMER_FREQ / OS_TICK_FREQ)) & 0xFF;
  TMR2RH  = (0x10000 - (OS_TIMER_FREQ / OS_TICK_FREQ)) >> 8;
  TMR2RLL = (0x10000 - (OS_TIMER_FREQ / OS_TICK_FREQ)) & 0xFF;
  TMR2RLH = (0x10000 - (OS_TIMER_FREQ / OS_TICK_FREQ)) >> 8;
  IE     |= (1u << IE_ET2_BIT);     // Enable timer interrupt
  TMR2CN |= (1u << TMR2CN_TR2_BIT); // Start timer
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  OS_COM_Init();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {  // Idle loop: No task is ready to execute
  while (1) {         // Nothing to do ... wait for interrupt
  }
}

/*********************************************************************
*
*       OS_BSP_GetCycles()
*
*  Function description
*    Returns the current system time in timer cycles as a 32-bit value.
*
*  Return value
*    The current system time in timer cycles as a 32-bit integer value.
*/
OS_U32 OS_BSP_GetCycles(void) {
  OS_U32 TickCnt;
  OS_U32 Cycles;

  OS_INT_IncDI();
  TickCnt = OS_TIME_GetTicks32();
  Cycles  = (OS_U32)TMR2RL;
  Cycles |= (OS_U32)TMR2RH << 8;
  //
  // Check if timer interrupt is pending
  //
  if ((TMR2CN & TMR2_TF2H_MASK) != 0u) {
    TickCnt++;
    Cycles  = (OS_U32)TMR2RL;
    Cycles |= (OS_U32)TMR2RH << 8;
  }
  OS_INT_DecRI();

  return ((TickCnt * (OS_TIMER_FREQ / OS_TICK_FREQ)) + Cycles);
}

/*********************************************************************
*
*       OS_BSP_ConvertCycles2us()
*
*  Function description
*    Converts timer cycles into microseconds.
*
*  Parameters
*    Cycles: Timer cycles
*
*  Return value
*    The converted value in microseconds.
*/
OS_U32 OS_BSP_ConvertCycles2us(OS_U32 Cycles) {
  return ((Cycles * 10u) / (OS_TIMER_FREQ / 100000u));
}

/*********************************************************************
*
*       OS_BSP_Convertus2Cycles()
*
*  Function description
*    Converts microseconds into timer cycles.
*
*  Parameters
*    us: Microseconds
*
*  Return value
*    The converted value in timer cycles
*/
OS_U32 OS_BSP_Convertus2Cycles(OS_U32 us) {
  return ((us * (OS_TIMER_FREQ / 100000u)) / 10u);
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
/*********************************************************************
*
*       OS_UART_interrupt()
*
*  Function description
*    Uart exception handler
*/
#pragma vector = 0x23
__interrupt void OS_UART_interrupt (void);
__interrupt void OS_UART_interrupt (void) {
  OS_U8 c;
  OS_INT_EnterNestable();
  if (SCON0 & UART0_RI0_MASK) {
    SCON0 &= ~UART0_RI0_MASK;
    c = SBUF0;
    OS_COM_OnRx(c);
  }
  if (SCON0 & UART0_TI0_MASK) {
    SCON0 &= ~UART0_TI0_MASK;
    OS_COM_OnTx();
  }
  OS_INT_LeaveNestable();
}

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
  SBUF0 = c;
}

/*********************************************************************
*
*       OS_COM_Init()
*
*  Function description
*    Initializes the uart.
*
*/
void OS_COM_Init(void) {
  //
  //  Set port pins for uart usage
  //
  P0MDOUT |= 0x10;            // Enable UTX as push-pull output
  XBR0     = 0x01;            // Enable UART on P0.4(TX) and P0.5(RX)
  XBR2     = 0x40;            // Enable crossbar and weak pull-ups
  //
  // Configure uart0
  //
  SCON0 = UART0_REN0_MASK;    // Uart in mode 1 (8 bit), REN=1
  //
  // Enable timer1 for generating baudrate
  //
  TH1    = (OS_U8)(0xFFFF - (OS_UART_FREQ/OS_BAUDRATE/2/48));
  CKCON &= ~0x0B;              // T1M = 0; SCA1:0 = 10
  CKCON |=  0x02;
  TL1    = TH1;                // Init timer1
  TMOD  &= ~0xf0;              // TMOD: timer 1 in 8-bit autoreload
  TMOD  |=  0x20;
  TCON  |= 0x40;               // Start timer1
  IE    |= IE_ES0_MASK;        // Enable uart0 interrupt
}
#endif  // (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)

/*********************************************************************
*
*       __low_level_init()
*
*  Function description
*    Called from IAR startup code.
*    Used to initialize hardware like PLL as early as possible.
*/
#ifdef __cplusplus
extern "C" {
#endif
__root char __low_level_init(void);    // Avoid "no prototype" warning
#ifdef __cplusplus
  }
#endif
__root char __low_level_init(void) {
  //
  // Disable watchdog
  //
  PCA0MD &= ~(1u << WDTE_BIT);
  //
  // Initialize clocks
  //
  OSCICN |= 0x80;                  // Enable the precision internal osc.
  while ((OSCICN & 0x40u) == 0u) { // Wait until clock is stable
  }
  CKCON   |= 0x30;                 // Timer2 uses sysclock
  RSTSRC |= 0x06;                  // Enable missing clock detector
  CLKSEL  = 0x00;                  // Switch to precision Internal Oscillator and set system clock is divided to 1
  while ((CLKSEL & 0x80u) == 0u) { // Wait until clock divide setting has been applied to the system clock.
  }

  return 1;                        // Always return 1 to enable segment initilization !
}

/*************************** End of file ****************************/
